#!/usr/bin/env python3

"""
 * Copyright (C) 2018 Arm Ltd
 * Author: Ash Wilding (ash.wilding@arm.com)
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.
"""

from sys import version_info as PYVER
assert (PYVER[0], PYVER[1]) >= (3,5), "script requires Python 3.5+"

import argparse
import bz2
import collections, ctypes
import gzip
import hashlib
import logging
import math
import os
import platform
import re
import shutil, subprocess, sys
import tarfile, threading, time
import urllib.error, urllib.request
import zipfile

HOST = platform.system()
assert HOST in ["Windows", "Linux"], "script requires Windows or Linux"
if HOST=="Windows":
    import ctypes

ARMPLATDB = {
  "arm": {
    "rel": "18.10",
    "vsn": "0.2.2",
    "manifests": "https://git.linaro.org/landing-teams/working/arm/manifest",
  },
  "linaro": {
    "url": "releases.linaro.org",
  },
  "p": {
    "name": "Platforms",
    "board": {
      "name": "Development boards",
      "juno": {
        "name": "Juno",
        "64b": {
          "name": "Juno with 64-bit software stack",
          "pdir": "juno",
          "deps": [
            "dl.tool.gcc.a64", "dl.tool.gcc.a32",
          ],          
          "android": {
            "rel": "18.04",
            "codename": "Oreo",
            "rootfs": "juno.img.bz2",
            "url": "{linaro.url}/members/arm/android/juno/{rel}",
            "ramdisk": {
              "dir": "prebuilts/android/juno",
            },
          },
          "oe": {
            "rel": "17.01",
            "vsn": "5.2_20170127-761",
            "url": "{linaro.url}/openembedded/juno-lsk/{rel}",
            "md5name": "{dl.img.oe.name}.md5",
          },
          "k": [
            "k.ack", "k.latest"
          ],
          "fs": [
            "fs.busybox", "fs.oe.mini", "fs.oe.lamp", "fs.android",
          ],
          "fw": [
            "fw.edkii",
          ],
          "pb": [
            "pb.ack.android",
            "pb.latest.busybox",
            "pb.latest.oe.mini",
            "pb.latest.oe.lamp",
            "pb.edkii",
          ],
        },
        "legacy": {
          "name": "Juno with legacy 32-bit software stack",
          "pdir": "juno32",
          "deps": [
            "dl.tool.gcc.a32",
          ],  
          "oe": {
            "rel": "15.07",
            "vsn": "4.9_20150725-725",
            "url": "{linaro.url}/openembedded/vexpress-lsk/{rel}",
          },
          "k": [
            "k.ack", "k.latest",
          ],
          "fs": [
            "fs.busybox", "fs.oe.alip",
          ],
          "fw": "null",
          "pb": [
            "pb.ack.busybox",
            "pb.latest.busybox",
            "pb.latest.oe.alip",
          ],
        },
      },
      "unsup": {
        "name": "Unsupported",
        "tc2": {
          "name": "TC2 with legacy 32-bit software stack",
          "pdir": "tc2",
          "deps": [
            "dl.tool.gcc.a32",
          ],
          "android": {
            "rel": "6.0-15.11",
            "codename": "Marshmallow",
            "rootfs": "vexpress.img.bz2",
            "url": "{linaro.url}/android/reference-lcr/vexpress/{rel}",
            "ramdisk": {
              "dir": "prebuilts/android/tc2",
            },
          },
          "oe": "#{p.board.juno.legacy.oe}",
          "k": "#{p.board.juno.legacy.k}",
          "fs": [
            "fs.busybox", "fs.oe.alip", "fs.android",
          ],
          "fw": [
            "fw.edkii",
          ],
          "pb": "null",
        },
      },
    },
    "fvp": {
      "name": "Fixed Virtual Platforms (FVPs)",
      "v8a": {
        "name": "Armv8 architecture",
        "pdir": "fvp",
        "base": {
          "name": "Armv8-A Base Platform",
          "descr": "11.3.30+ (Rev C)",
          "64b": {
            "name": "Armv8-A Base Platform with 64-bit software stack",
            "deps": [
              "dl.tool.gcc.a64", "dl.tool.gcc.a32",
            ],  
            "descr": "null",
            "android": {
              "rel": "7.0-16.10",
              "codename": "Nougat",
              "rootfs": "fvp.img.bz2",
              "url": "{linaro.url}/android/reference-lcr/fvp/{rel}",
              "ramdisk": {
                "dir": "prebuilts/android/fvp",
              },
            },
            "oe": {
              "rel": "15.09",
              "vsn": "4.9_20150912-729",
              "url": "{linaro.url}/openembedded/juno-lsk/{rel}",
              "md5name": "MD5SUMS.txt",
            },
            "k": [
              "k.ack", "k.latest",
            ],
            "fs": [
              "fs.busybox", "fs.oe.mini", "fs.oe.lamp", "fs.android",
            ],
            "fw": [
              "fw.edkii",
            ],
            "pb": [
              #"pb.ack.android.debug",
              "pb.ack.busybox",
              "pb.latest.busybox",
              #"pb.latest.oe.mini.debug",
              #"pb.latest.oe.lamp.debug",
              "pb.edkii",
            ],
          },
          "legacy": {
            "name": "Armv8-A Base Platform with legacy 32-bit software stack",
            "pdir": "fvp32",
            "deps": [
              "dl.tool.gcc.a32",
            ],
            "descr": "null",
            "oe": "#{p.board.juno.legacy.oe}",
            "k": "#{p.board.juno.legacy.k}",
            "fs": "#{p.board.juno.legacy.fs}",
            "fw": "null",
            "pb": "#{p.board.juno.legacy.pb}",
          },
        },
        "fndn": {
          "name": "Armv8-A Foundation Platform",
          "descr": "11.3.30+",
          "64b": {
            "name": "Armv8-A Foundation Platform with 64-bit software stack",
            "deps": [
              "dl.tool.gcc.a64", "dl.tool.gcc.a32",
            ],  
            "descr": "null",
            "android": "#{p.fvp.v8a.base.64b.android}",
            "oe": "#{p.fvp.v8a.base.64b.oe}",
            "k": "#{p.fvp.v8a.base.64b.k}",
            "fs": "#{p.fvp.v8a.base.64b.fs}",
            "fw": "#{p.fvp.v8a.base.64b.fw}",
            "pb": "#{p.fvp.v8a.base.64b.pb}",
          },
        },
      },
      "sg": {
        "name": "System Guidance",
        "deps": [
          "dl.tool.gcc.a64", "dl.tool.gcc.a32", "dl.tool.gcc.scp",
        ],  
        "i": {
          "name": "System Guidance for Infrastructure (SGI)",
          "575": {
            "name": "SGI-575",
            "pdir": "sgi575",
            "linux": {
              "vsn": "4.18",
            },
            "fedora": {
              "rel": "27",
              "vsn": "27-1.6",
            },
            "k": [
              "k.mainline",
            ],
            "fs": [
              "fs.busybox", "fs.fedora",
            ],
            "pb": [
              "pb.latest.busybox.edkii",
            ],
            "fw": "null",
          },
        },
        "m": {
          "name": "System Guidance for Mobile (SGM)",
          "775": {
            "name": "SGM-775",
            "pdir": "sgm775",
            "android": {
              "rel": "(built from source)",
              "codename": "Oreo",
            },
            "k": [
              "k.ack.sgm775",
              "k.latest.sgm775",
            ],
            "fs": [
              "fs.android.bfs",
              "fs.busybox",
            ],
            "pb": [
              "pb.ack.busybox",
              "pb.ack.android.bfs",
            ],
            "fw": "null",
          },
        },
      },
    },
    "all": [
      "p.board.juno.64b", "p.board.juno.legacy", "p.board.unsup.tc2",
      "p.fvp.v8a.base.64b", "p.fvp.v8a.base.legacy", "p.fvp.v8a.fndn.64b",
      "p.fvp.sg.i.575", "p.fvp.sg.m.775",
    ],
  },
  "k": {
    "name": "Linux kernel + filesystem",
    "ack": {
      "name": "ack-4.9-armlt",
      "code": "ack",
      "manifest": "pinned-{code}.xml",
      "descr": "Android Common Kernel",
      "fs": [
        "fs.busybox", "fs.oe.alip", "fs.oe.mini", "fs.oe.lamp", "fs.android"
      ],
      "sgm775": {
        "manifest": "sgm775-android.xml",
        "fs": [
          "fs.android.bfs",
        ],
      },
    },
    "latest": {
      "name0": "latest-armlt",
      "name": "{name0}",
      "code": "latest",
      "manifest": "pinned-{code}.xml",
      "descr": "4.17 commit 9b2769768c",
      "fs": [
        "fs.busybox", "fs.oe.alip", "fs.oe.mini", "fs.oe.lamp",
      ],
      "sgm775": {
        "name": "{name0} for {@.name}",
        "code": "{p.fvp.sg.m.775.pdir}",
        "manifest": "{code}.xml",
        "fs": [
          "fs.busybox",
        ],
      },
    },
    "mainline": {
      "name": "Mainline {@.linux.vsn}",
      "code": "{@.pdir}",
      "manifest": "{code}.xml",
      "fs": [
        "fs.busybox", "fs.fedora",
      ],
    },
  },
  "fs": {
    "android": {
      "name": "Android",
      "script": "android",
      "descr": "{@.android.codename} {@.android.rel}",
      "deps": [
        "dl.img.android.rootfs", "dl.img.android.ramdisk"
      ],
      "bfs": {
        "deps": "null",
      },
    },
    "busybox": {
      "name": "BusyBox",
      "script": "busybox",
      "descr": "commit 111cdcf295",
    },
    "oe": {
      "name": "OpenEmbedded",
      "script": "oe",
      "descr": "{@.oe.rel}",
      "alip": {
        "name": "OpenEmbedded ALIP",
        "deps": [
          "dl.img.oe.alip",
        ],
      },
      "mini": {
        "name": "OpenEmbedded Minimal",
        "deps": [
          "dl.img.oe.mini"
        ],
      },
      "lamp": {
        "name": "OpenEmbedded LAMP",
        "deps": [
          "dl.img.oe.lamp"
        ],
      },
    },
    "fedora": {
      "name": "Fedora Server",
      "script": "busybox",
      "descr": "{@.fedora.vsn}",
      "deps": [
        "dl.img.fedora",
      ],
    },
  },
  "fw": {
    "name": "Firmware only",
    "edkii" : {
      "name": "EDK II UEFI",
      "stubfs": "uefi",
      "manifest": "pinned-{stubfs}.xml",
    },
  },
  "pb": {
    "ack": {
      "name0": "{k.ack.name}",
      "android": {
        "name": "{name0} + {fs.android.name} {fs.android.descr}",
        "deps": [
          "dl.img.android.rootfs",
          "dl.archive.ack.android",
        ],
        "debug": {
          "deps": [
            "dl.img.android.rootfs",
            "dl.archive.ack.android",
            "dl.archive.ack.android.debug",
          ],
        },
        "bfs": {
          "name": "{name0} + {fs.android.name} {@.android.codename} (snapshot)",
          "deps": [
            "dl.archive.ack.android",
          ],
        },
      },
      "busybox": {
        "name": "{name0} + {fs.busybox.name} {fs.busybox.descr}",
        "deps": [
          "dl.archive.ack.busybox"
        ],
      },
    },
    "latest": {
      "name0": "{k.latest.name}",
      "busybox": {
        "name": "{name0} + {fs.busybox.name} {fs.busybox.descr}",
        "deps": [
          "dl.archive.latest.busybox"
        ],
        "edkii": {
          "deps": [
            "dl.archive.latest.busybox.edkii"
          ],
        },
      },
      "oe": {
        "alip": {
          "name": "{name0} + {fs.oe.alip.name} {fs.oe.descr}",
          "deps": [
            "dl.img.oe.alip",
            "dl.archive.latest.oe",
          ],
        },
        "mini": {
          "name": "{name0} + {fs.oe.mini.name} {fs.oe.descr}",
          "deps": [
            "dl.img.oe.mini",
            "dl.archive.latest.oe",
          ],
          "debug": {
            "deps": [
              "dl.img.oe.mini",
              "dl.archive.latest.oe",
              "dl.archive.latest.oe.debug",
            ],
          },
        },
        "lamp": {
          "name": "{name0} + {fs.oe.lamp.name} {fs.oe.descr}",
          "deps": [
            "dl.img.oe.lamp",
            "dl.archive.latest.oe",
          ],
          "debug": {
            "deps": [
              "dl.img.oe.lamp",
              "dl.archive.latest.oe",
              "dl.archive.latest.oe.debug",
            ],
          },
        },
      },
    },
    "edkii": {
      "name": "{fw.edkii.name}",
      "deps": [
        "dl.archive.edkii",
      ],
    },
  },
  "dl": {
    "tool": {
      "gcc": {
        "rel": "6.2-2016.11",
        "vsn": "6.2.1-2016.11",
        "url": "http://{linaro.url}/components/toolchain/binaries/{rel}/{tuple}",
        "name": "gcc-linaro-{vsn}-x86_64_{tuple}.tar.xz",
        "md5name": "{name}.asc",
        "dir": "tools/gcc",
        "a32": {
          "tuple": "arm-linux-gnueabihf",
        },
        "a64": {
          "tuple": "aarch64-linux-gnu",
        },
        "scp": {
          "rel": "5.0/5-2016-q3-update",
          "vsn": "5_4-2016q3-20160926",
          "url": "https://launchpad.net/gcc-arm-embedded/{rel}/+download",
          "tuple": "arm-none-eabi",
          "name": "gcc-{tuple}-{vsn}-linux.tar.bz2",
          "md5name": "{name}/+md5",
        },
        "both": [
          "dl.tool.gcc.a32", "dl.tool.gcc.a64",
        ],
      },
      "repo": {
        "name": "repo",
        "dir": "tools",
        "url": "storage.googleapis.com/git-repo-downloads",
      },
    },
    "img": {
      "dir": ".",
      "android": {
        "url": "{@.android.url}",
        "md5name": "MD5SUMS",
        "rootfs": {
          "name": "{@.android.rootfs}",
        },
        "ramdisk": {
          "name": "ramdisk.img",
          "dir": "{@.android.ramdisk.dir}"
        },
      },
      "oe": {
        "url": "{@.oe.url}",
        "name": "{name0}-openembedded_{name1}-{name2}-gcc-{@.oe.vsn}.img.gz",
        "alip": {
          "name0": "lsk-vexpress",
          "name1": "alip",
          "name2": "armv7a",
          "md5name":"null",
        },
        "mini": {
          "name0": "lt-vexpress64",
          "name1": "minimal",
          "name2": "armv8",
          "md5name": "{@.oe.md5name}",
        },
        "lamp": {
          "name0": "lt-vexpress64",
          "name1": "lamp",
          "name2": "armv8",
          "md5name": "{@.oe.md5name}",
        },
      },
      "fedora": {
        "url": "https://dl.fedoraproject.org/pub/fedora-secondary/releases" \
            + "/{@.fedora.rel}/Server/aarch64/iso/",
        "name": "Fedora-Server-dvd-aarch64-{@.fedora.vsn}.iso",
      },
    },
    "archive": {
      "url": "{linaro.url}/members/arm/platforms/{arm.rel}",
      "md5name": "MD5SUMS",
      "dir": "{basename}",
      "name0": "{@.pdir}",
      "name3": "uboot",
      "basename": "{name0}-{name1}-{name2}-{name3}",
      "name": "{basename}.{fmt}",
      "fmt": "zip",
      "ack": {
        "name1": "{k.ack.code}",
        "android": {
          "name2": "{fs.android.script}",
          "debug": {
            "name3": "debug",
            "fmt": "tar.xz",
          },
        },
        "busybox": {
          "name2": "{fs.busybox.script}",
        },
      },
      "latest": {
        "name1": "{k.latest.code}",
        "busybox": {
          "name2": "{fs.busybox.script}",
          "edkii": {
            "name3": "{fw.edkii.stubfs}",
          },
        },
        "oe": {
          "name2": "{fs.oe.script}",
          "debug": {
            "name3": "debug",
            "fmt": "tar.xz",
          },
        },
      },
      "edkii": {
        "basename": "{name0}-{fw.edkii.stubfs}",
      },
    },
  },
  "apt": {
    "deps": [
      "bison", "dh-autoreconf", "device-tree-compiler", "expect", "flex", 
      "fuseext2", "g++", "gcc", "git", "libssl-dev", "openjdk-8-jdk",
      "python-crypto", "python-wand", "uuid-dev", "xterm",
    ],
  },
}


class Database(dict):
    def lookup( self, key, plat=None, noneAllowed=False ):
        log.debug("lookup {} (plat={}, nA={})".format(key, plat, noneAllowed))
        def subPlat(k):
            try:
                return k.replace("@", plat) if "@" in k else k
            except TypeError:
                script.abort("lookup key={} with invalid plat={}"
                             .format(key, plat))
        # We assimilate keys into d at each level of lookup, allowing for keys
        # to be inherited and overridden by later levels.
        d = {}
        assimilate = lambda src: [d.update({k:v}) for k,v in src.items()]
        assimilate(self)
        # Perform recursive lookup
        (lookupLvl, item) = (d, None)
        for k in subPlat(key).split("."):
            item = lookupLvl[k] if k in lookupLvl else d[k] if k in d else None
            if item is None:
                break
            # Handle cross-references
            if isinstance(item, str) and item.startswith("#"):
                if item.count("{")==item.count("}")==1:
                    (l, r) = (item.find("{"), item.find("}"))
                    item = Database(d).lookup(item[l+1:r], plat)
            # Prepare for next level of lookup
            if isinstance(item, dict):
                item = Database(item)
                lookupLvl = item
                assimilate(lookupLvl)
        # String items have special behaviour
        if isinstance(item, str):
            if "null"==item:
                item = None
            else:
                item = subPlat(item)
                if not item.count("{")==item.count("}"):
                    script.abort(("lookup of key={} with plat={} gives item={}"
                                  " with imbalanced number of {{ and }}")
                                  .format(key, plat, item))
                while "{" in item:
                    (l, r) = (item.find("{"), item.find("}"))
                    item = item.replace(item[l:r+1],
                                        Database(d).lookup(item[l+1:r], plat))
        if item is None and not noneAllowed:
            script.abort(("lookup of {} (plat={}) returns None but noneAllowed"
                          " is False").format(key, plat))
        return item


    def multilookup( self, root, keys, plat=None, noneAllowed=False ):
        return [self.lookup(root+"."+k, plat, noneAllowed) for k in keys]


def dblu( key, plat=None, noneAllowed=False ):
    return Database(ARMPLATDB).lookup(key, plat, noneAllowed)


def dblum( root, keys, plat=None, noneAllowed=False ):
    return Database(ARMPLATDB).multilookup(root, keys, plat, noneAllowed)


class script:
    def init():
        # Parse arguments
        p = argparse.ArgumentParser()
        p.add_argument("-v", help="verbose info", action="count", default=0)
        p.add_argument("--qa_mode", help="for Arm internal QA purposes",
                       action="store_true")
        p.add_argument("--no_check_apt_deps", action="store_true",
                       help="do not check for APT package dependencies")
        args = p.parse_args()
        (script.v, script.qa_mode, script.no_check_apt_deps) = \
            (args.v, args.qa_mode, args.no_check_apt_deps)
        # Configure logging
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.DEBUG)
        console = logging.StreamHandler()
        # Logging level increases with number of -v on command line
        script.loglvl = {0:logging.WARNING, 1:logging.INFO, 2:logging.DEBUG} \
            [max(0, min(script.v, 2))]
        console.setLevel(script.loglvl)
        console.setFormatter(logging.Formatter("%(levelname)s: %(message)s"))
        logger.addHandler(console)
        # 2+ -v enables logging to file
        if script.v > 2:
            fh = logging.FileHandler("log.txt", "w")
            fh.setLevel(logging.DEBUG)
            fh.setFormatter(logging.Formatter("%(levelname)s: %(message)s"))
            logger.addHandler(fh)
        log.info("Arm Platforms {} workspace initialization script {}"
                 .format(dblu("arm.rel"), dblu("arm.vsn")))
        log.info("date is {}".format(time.strftime("%Y-%m-%d %H:%M:%S")))
        log.info("running on {} host".format(HOST))
        # Other setup
        script.aborts = []
        sh.init()


    def start_qa():
        script.qa_t0 = time.time()


    def end_qa( hard_aborted=False ):
        t = time.time() - script.qa_t0
        hrs = math.floor(t / 3600)
        mins = math.floor((t % 3600) / 60)
        secs = math.floor((t % 3600) % 60)
        logfn = log.info if len(script.aborts)==0 else log.error
        msg = "QA run {} after {}hrs {}mins {}secs" \
              .format("completed" if not hard_aborted else "HARD ABORTED",
                      hrs, mins, secs)
        wall = "#"*len(msg)
        logfn(wall)
        logfn(msg)
        logfn("Total number of aborts: {}".format(len(script.aborts)))
        logfn(wall)
        logfn("QA run result: {}".format(
            "SUCCESS" if len(script.aborts)==0 else "FAIL"))
        logfn(wall)
        return len(script.aborts)


    def abort( e=None, hard=True ):
        if e:
            log.error("")
            log.error(str(e))
        script.aborts.append(str(e) if e else "<<empty abort : see full log>>")
        do_exit = True
        if script.qa_mode:
            if hard:
                script.end_qa(hard_aborted=True)
            else:
                do_exit = False
        if do_exit:
            sys.exit(e.errno if hasattr(e, "errno") else -1)


class log:
    def debug( msg ): logging.getLogger(__name__).debug(msg)
    def info ( msg ): logging.getLogger(__name__).info (msg)
    def warn ( msg ): logging.getLogger(__name__).warn (msg)
    def error( msg ): logging.getLogger(__name__).error(msg)


class sh:
    def init():
        (srd, cwd) = (sh.fmtpath(sys.path[0])+"/", sh.fmtpath(os.getcwd())+"/")
        if HOST=="Linux":
            (srd, cwd) = ("/"+srd, "/"+cwd)
        sh.cwd = cwd if cwd==srd else \
            prompt("Current working dir differs from script resident dir.\n"
                   "## Please specify which to initialize as your workspace",
                   [
                     choice(cwd, descr="working", meta=cwd),
                     choice(srd, descr="script", meta=srd)
                   ]
            ).meta
        try:
            os.chdir(sh.cwd)
        except OSError as e:
            script.abort(e)
        sh.dld = sh.mkdir("downloads", hidden=True)+"/"
        sh.filename = __file__.rsplit("/" if HOST=="Linux" else "\\")[-1]
        sh.repod = ".repo/"
        check_empty_ws()


    def fmtpath(p):
        p = p.replace("\\", "/")
        while "//" in p:
            p = p.replace("//", "/")
        return p.strip().lstrip("/").rstrip("/")


    def _op( func, *paths, extra=None, silent=False ):
        paths = [sh.fmtpath(p) for p in paths]
        if not silent:
            log.info("{} {} {}".format(func, *paths, extra if extra else ""))
        try:
            return func(*paths) if not extra else func(*paths, extra)
        except OSError as e:
            script.abort(e)


    def mkdir( p, hidden=False ):
        if hidden:
            slash = p.rfind("/")
            p = "{}.{}".format(p[:slash+1], p[slash+1:])
        if not sh._op(os.path.isdir, p, silent=True):
            sh._op(os.makedirs, p)
        if hidden and HOST=="Windows":
            ret = sh._op(ctypes.windll.kernel32.SetFileAttributesW, p,
                         extra=0x2) # 0x2 == FILE_ATTRIBUTE_HIDDEN
            if 0==ret:
                script.abort(ctypes.WinError())
        return p


    def rmdir(p):
        if sh._op(os.path.isdir, p):
            sh._op(shutil.rmtree, p)


    def cp( src, dstdir ):
        sh._op(shutil.copy, src, dstdir)
        return "{}/{}".format(dstdir, src.split("/")[-1])


    def rm(p):
        if sh._op(os.path.isfile, p):
            sh._op(os.remove, p)


    def _std_extract( func, src, dstdir, *args, **kwargs ):
        (src, dstdir) = (sh.fmtpath(src), sh.fmtpath(dstdir))
        log.info("{} {} {}".format(func, src, dstdir))
        try:
            with func(src, *args, **kwargs) as f:
                f.extractall(dstdir)
        except OSError as e:
            script.abort(e)
        return dstdir


    def _bin_extract( func, src, dstdir, extn ):
        (src, dstdir) = (sh.fmtpath(src), sh.fmtpath(dstdir))
        dst = "{}/{}".format(dstdir, src.split("/")[-1][:-len(extn)])
        log.info("{} {} {}".format(func, src, dst))
        try:
            with func(src, "rb") as inf:
                with open(dst, "wb") as outf:
                    while True:
                        chunk = inf.read(0x1000)
                        if not chunk:
                            break
                        outf.write(chunk)
        except OSError as e:
            script.abort(e)
        return dst


    def _tarxf( src, dstdir ):
        return sh._std_extract(tarfile.open, src, dstdir, "r:xz", errorlevel=1)


    def _tarxjf( src, dstdir ):
        return sh._std_extract(tarfile.open, src, dstdir, "r:bz2", errorlevel=1)


    def _unzip( src, dstdir ):
        return sh._std_extract(zipfile.ZipFile, src, dstdir, "r")


    def _gunzip( src, dstdir ):
        return sh._bin_extract(gzip.open, src, dstdir, ".gz")


    def _bunzip2( src, dstdir ):
        return sh._bin_extract(bz2.BZ2File, src, dstdir, ".bz2")


    def extract_or_copy( src, dstdir ):
        sh.mkdir(dstdir)
        ends = lambda s: src.endswith(s)
        handler = sh._tarxf   if ends(".tar.xz") else \
                  sh._tarxjf  if ends(".tar.bz2") else \
                  sh._unzip   if ends(".zip") else \
                  sh._gunzip  if ends(".gz") and not ends(".tar.gz") else \
                  sh._bunzip2 if ends(".bz2") and not ends(".tar.bz2") else \
                  sh.cp
        return handler(src, dstdir)


    def md5sum(p):
        p = sh.fmtpath(p)
        log.info("md5sum "+p)
        md5 = hashlib.md5()
        try:
            with open(p, "rb") as f:
                while True:
                    block = f.read(md5.block_size)
                    if not block:
                        break
                    md5.update(block)
        except OSError as e:
            script.abort(e)
        md5 = md5.hexdigest()
        log.info("got hash "+md5)
        return md5


    def md5check( p, sumsp ):
        (p, sumsp) = (sh.fmtpath(p), sh.fmtpath(sumsp))
        (md5, name) = (sh.md5sum(p), p.split("/")[-1])
        log.debug("checking md5 of {} in {}".format(name, sumsp))
        try:
            with open(sumsp, "r") as sumsf:
                rstr = re.compile(r"([^\s]+)\s+"+re.escape(name))
                regex = re.search(rstr, sumsf.read())
        except OSError as e:
            script.abort(e)
        if not regex:
            script.abort("no md5 for {} in sumsfile {}"
                         .format(name, sumsp), hard=False)
            return False
        match = "MATCH" if (regex.group(1) == md5) else "MISMATCH"
        log.info("md5 {} {}".format(match, regex.group(1)))
        return (regex.group(1) == md5)


    def url2dld(url):
        p = url.lstrip("/").rstrip("/").replace(":","_").replace("/","_")
        while "__" in p:
            p = p.replace("__", "_")
        return "{}/{}".format(sh.dld, p)


    def wget( url, name, silent=False ):
        def si( num_bytes ):
            if num_bytes <= 0:
                return "0 B"
            suffixes = ["B", "KiB", "MiB", "GiB"]
            log = int(math.floor(math.log(num_bytes, 1024)))
            size = round(num_bytes / 1024**log, 2)
            return "{:.2f} {}".format(size, suffixes[log])
        dld = sh.url2dld(url)
        url = "{}/{}".format(url, name)
        log.info("wget "+url)
        if not url.startswith("http://") and not url.startswith("https://"):
            url = "http://"+url
            log.debug("url does not specify protocol, defaulting to http")
        try:
            r = urllib.request.urlopen(url)
            log.debug("connected to server")
            try:
                endln = " / "+si(int(r.getheader("Content-Length").strip()))
            except ValueError():
                endln = ""
                log.debug("server does not support Content-Length header")
            sh.mkdir(dld)
            dst = "{}/{}".format(dld, name.split("/")[-1])
            log.info("opening dst file "+dst)
            with open(dst, "wb") as f:
                progress = 0
                eraseln = "\r{}\r" \
                          .format(" "*(shutil.get_terminal_size().columns-1))
                while True:
                    chunk = r.read(0x100000)
                    if not chunk:
                        break
                    f.write(chunk)
                    progress += len(chunk)
                    if not silent:
                        print("{}Fetching {}: {}{}".format(eraseln, name,
                              si(progress), endln), end="")
                    sys.stdout.flush()
                if not silent:
                    print() ### for \n\r
            log.info("successfully fetched "+si(progress))
            return True
        except (OSError, urllib.error.HTTPError) as e:
            script.abort(e, hard=False)
            return False


    def fetch( key, plat=None, force_fresh=False ):
        (url, name, md5name, dstdir) =  \
            dblum(key, ["url","name","md5name","dir"], plat, noneAllowed=True)
        dld = sh.url2dld(url)
        dlfile = "{}/{}".format(dld, name.split("/")[-1])
        if md5name:
            md5file = "{}/{}".format(dld, md5name.split("/")[-1])
            sh.wget(url, md5name, silent=True)
        already_fetched = False
        if sh._op(os.path.isfile, dlfile, silent=True):
            if md5name and sh._op(os.path.isfile, md5file, silent=True):
                if not force_fresh and sh.md5check(dlfile, md5file):
                    already_fetched = True
                    print("Already fetched "+name)
                    sys.stdout.flush()
        if not already_fetched:
            sh.wget(url, name)
            if md5name and sh._op(os.path.isfile, md5file, silent=True):
                if not sh.md5check(dlfile, md5file):
                    script.abort("MD5 mismatch after fetching {} (sumf {})"
                                .format(dlfile, md5file, hard=False))
        print("Extracting {}...".format(name))
        return sh.extract_or_copy(dlfile, dstdir)


    def reposync( manifest, force_fresh=False ):
        log.info("Attempting to sync "+manifest)
        if force_fresh:
            sh.rmdir(".repo")
        repo = sh.fetch("dl.tool.repo")
        def call_repo(argstr):
            sp = subprocess
            log.debug("calling repo "+argstr)
            proc = sp.Popen(["unbuffer", "python2", repo] + argstr.split(" "),
                            stdout=sp.PIPE, stderr=sp.STDOUT, bufsize=0,
                            universal_newlines=True)
            ln = ""
            while proc.returncode is None:
                ln += os.read(proc.stdout.fileno(), 1).decode("utf-8")
                if ln.endswith("\r") or ln.endswith("\n"):
                    log.info(ln.rstrip("\n"))
                    ln = ""
                proc.poll()
            return proc.returncode
        def init():
            print("\nInitialising repo")
            if not 0==call_repo("init -b {} -m {} -u {}" \
              .format(dblu("arm.rel"), manifest, dblu("arm.manifests"))):
                script.abort("failed to initialise repo")
        def sync():
            print("Syncing repo")
            print("NOTE: this can take a very long time, please be patient")
            print("NOTE: run script with -v for detailed repo progress info")
            if not 0==call_repo("sync -j8 --force-sync"):
                script.abort("failed to sync repo", hard=False)
        init()
        sync()


    def call( args, err2out=False, pipe2sh=False ):
        log.info("calling {}".format(" ".join(args)))
        sp = subprocess
        stdout = None if pipe2sh else sp.PIPE
        stderr = sp.STDOUT if err2out else None if pipe2sh else sp.PIPE
        with sp.Popen(args, stdout=stdout, stderr=stderr) as p:
            (out, err) = p.communicate()
        strip = lambda s: s.decode("utf-8").strip() if s else ""
        return (p.returncode, strip(out), None if err2out else strip(err))


class choice():
    def __init__( self, name, meta=None, descr=None, disabled=None,
                  children=None ):
        (self.name, self.meta, self.descr, self.disabled, self.children) = \
            (name, meta, descr, disabled, [])
        if children:
            [self.add(c) for c in children]


    def add( self, child ):
        log.debug("append child "+child.meta)
        self.children.append(child)


    def tree( keylist, root, plat=None, gen_root=True ):
        if gen_root:
            root = choice("<root>", meta=root)
        log.debug("tree({})".format(root.meta))
        keylist = list(filter(lambda k: k.startswith(root.meta), keylist))
        if len(keylist)==1 and keylist[0]==root.meta:
            log.debug("{} is leaf node".format(root.meta))
            return None
        keys = []
        for k in keylist:
            log.debug(">>>> "+k)
            key = k[len(root.meta)+1:]
            if "." in key:
                key = key[:key.find(".")]
            keys.append(key)
        keys = ["{}.{}".format(root.meta, k) for k in sorted(list(set(keys)))]
        if len(keys) > 0:
            log.debug("generating child nodes")
            for k in keys:
                log.debug("k="+k)
                root.add(choice(
                    dblu(k+".name", plat), meta=k,
                    descr=dblu(k+".descr", plat, noneAllowed=True)))
            log.debug("crawling child nodes")
            for c in root.children:
                choice.tree(
                    keylist,root.children[root.children.index(c)],plat,False)
        return root


def prompt( title, choices ):
    pad = 3 + max([len(c.name) for c in choices])
    def fmt_choice( c ):
        descr = "-- <DISABLED>" if c.disabled else \
                "-- {}".format(c.descr) if c.descr else ""
        return "{:>2}) {:{}}{}".format(choices.index(c)+1, c.name, pad, descr)
    msg = "\n\n## {}:\n\n{}\n\n> " \
          .format(title, "\n".join(map(fmt_choice, choices)))
    for ln in msg.splitlines():
        if not ln.startswith(">"):
            log.debug(ln)
    while True:
        try:
            i = int(input(msg))
            if (i < 1) or (i > len(choices)):
                raise ValueError()
            i = i - 1
            if choices[i].disabled:
                print("Not available: {}".format(choices[i].descr))
            else:
                break
        except ValueError:
            print("Expected number in range [1..{}] inclusive.\n"
                  .format(str(len(choices))))
    log.debug("> {} (meta={})".format(i+1, choices[i].meta))
    return choices[i]


def tree_prompt( title, root ):
    log.debug("tree_prompt "+title)
    log.debug("root has {} children".format(len(root.children)))
    while root.children:
        if len(root.children)==1:
            root = root.children[0]
            log.warn("only avail. option is "+root.name)
        else:
            root = prompt(title, root.children)
    return root


def check_apt_deps():
    if script.no_check_apt_deps:
        log.debug("skipping APT dependency check")
        return
    try:
        import apt
    except:
        log.error("cannot check package dependencies (failed to import apt)")
        log.error("the following apt packages are required:")
        log.error("")
        [log.error(" "+d) for d in dblu("apt.deps")]
        log.error("")
        log.error("please install python3-apt or check package dependencies")
        log.error("manually and re-run the script with `--no_check_apt_deps'")
        script.abort()
    log.debug("checking APT package dependencies")
    cache = apt.Cache()
    missing = filter(lambda d: not cache[d].is_installed, dblu("apt.deps"))
    missing = list(missing)
    if missing:
        log.error("the following apt packages are missing:")
        log.error("")
        [log.error(" "+d) for d in missing]
        log.error("")
        log.error("please install these packages using:")
        log.error("")
        log.error("sudo apt-get install {}".format(" ".join(missing)))
        script.abort()


def check_sys_gcc():
    log.debug("checking system gcc")
    (_, out, _) = sh.call(["gcc", "-dumpversion"])
    ver = tuple([int(i) for i in out.split(".")])
    if ver < (5,4,0):
        script.abort(("detected system native gcc version {}.{}.{}, please "
                      "upgrade to version 5.4.0 or newer")
                      .format(ver[0], ver[1], ver[2]))


def check_git_config():
    log.debug("checking git config")
    (_, cfg, _)  = sh.call(["git", "config", "-l"])
    for setting in ["user.name", "user.email", "color.diff"]:
        if not setting in cfg:
            log.error("git is not correctly configured")
            log.error("please ensure the following git configs are set:")
            log.error("")
            log.error("git config --global user.name \"Joe Bloggs\"")
            log.error("git config --global user.email \"jb@example.com\"")
            log.error("git config --global color.diff \"auto\"")
            script.abort()


def check_empty_ws():
    filelist = []
    for (_, dirs, files) in os.walk(sh.cwd):
        dirs = [d+"/" for d in dirs]
        filelist += sorted(files + dirs)
        break  # don't recursively walk
    log.debug("found the following files in workspace directory:")
    log.debug(filelist)
    for f in [sh.filename, sh.dld, sh.repod, "log.txt"]:
        if f in filelist:
            filelist.remove(f)
    if len(filelist) > 0:
        msg = ("Expected empty workspace but found the following files & "
               "folders:\n\n - {}\n\n## Delete these files & folders and "
               "proceed?").format("\n - ".join(filelist))
        if not prompt(msg, [choice("Yes", True), choice("No", False)]).meta:
            sys.exit(0)
        [sh.rmdir(f) if f.endswith("/") else sh.rm(f) for f in filelist]


class config:
    def query():
        while not config._choose():
            pass


    def sync():
        print("\nFetching and extracting dependencies...\n")
        for d in config.deps:
            sh.fetch(d, plat=config.p.meta)
        if config.manifest:
            sh.reposync(config.manifest)
        if config.ws.meta=="bfs":
            for (_, dirs, _) in os.walk("build-scripts/platforms"):
                break  # don't recursively walk
            for d in dirs:
                if not (d=="common" or d==dblu(config.p.meta+".pdir")):
                    sh.rmdir("build-scripts/platforms/"+d)
            for (_, _, files) in os.walk("build-scripts/filesystems"):
                break  # don't recursively walk
            preserve = dblu(config.fs.meta+".script") if config.env.meta=="k" \
                  else dblu(config.fw.meta+".stubfs")
            for f in files:
                if not f==preserve:
                    sh.rm("build-scripts/filesystems/"+f)
        print("\nWorkspace initialised.")
        if config.ws.meta=="bfs":
            print("\nTo build:\n")
            print("    chmod a+x <workspace>/build-scripts/build-all.sh")
            print("    <workspace>/build-scripts/build-all.sh all")
            print("\nResulting binaries will be placed in:")
            print("\n    <workspace>/output/{}-{}/".format(
                      dblu("@.pdir", config.p.meta),
                      dblu(config.fs.meta+".script" if config.env.meta=="k" \
                          else config.fw.meta+".stubfs", config.p.meta)))
        print("\nFor more information, tutorials, FAQs, and discussions, see:")
        print("\n    https://www.community.arm.com/tools/dev-platforms/")
        print("\nThank you for using the Arm Platforms workspace script.")


    def _choose():
        config.cfg = []
        config.deps = []
        config.manifest = None
        config._add_cfg("Workspace", sh.cwd)
        config._choose_p()
        config._choose_ws()
        if config.ws.meta=="bfs":
            check_apt_deps()
            check_git_config()
            check_sys_gcc()
            config._add_deps(config.p.meta)
            config._choose_env()
            if config.env.meta=="k":
                config._choose_k()
                config._choose_fs()
            elif config.env.meta=="fw":
                config._choose_fw()
        elif config.ws.meta=="pbc":
            config._choose_pb()
        msg = "Your chosen configuration is shown below:\n"
        pad = [max([len(row[n]) for row in config.cfg]) for n in [0,1]]
        wall = "\n    +-"+(pad[0]*"-")+"-+-"+(pad[1]*"-")+"-+"
        msg += wall
        for row in config.cfg:
            msg += "\n    | {:<{:}} | {:<{:}} |" \
                      .format(row[0], pad[0], row[1], pad[1])
        msg += wall
        msg += "\n\n## Proceed with this configuration?"
        return prompt(msg, [choice("Yes", True), choice("No", False)]).meta


    def _add_cfg( key, val ):
        config.cfg.append((key, val))


    def _add_deps(key):
        deps = dblu(key+".deps", config.p.meta, noneAllowed=True)
        if deps:
            for d in deps:
                config.deps.append(d)


    def _choose_p():
        log.debug("Building platform choice tree")
        def crawl(root):
            log.debug("crawl({})".format(root.meta))
            num_en_children = 0
            if root.children:
                log.debug("{} is non-leaf".format(root.meta))
                [crawl(c) for c in root.children]
                num_en_children = sum([not p.disabled for p in root.children])
                log.debug("{} has {} enabled children"
                          .format(root.meta, num_en_children))
            pbs = dblu(root.meta+".pb", noneAllowed=True)
            num_pbs = len(pbs) if pbs else 0
            if not root.children:
                log.debug("{} has {} prebuilts".format(root.meta, num_pbs))
            can_init_ws = True
            if num_pbs==0 and not HOST=="Linux":
                can_init_ws = False
                if not root.children:
                    log.debug("not running on Linux")
            log.debug("can_init_ws {}? {}".format(root.meta, can_init_ws))
            if (root.children and num_en_children==0) or not can_init_ws:
                log.debug("disabling node {}".format(root.meta))
                root.disabled = True
                root.descr = "can only build from source for this platform, " \
                             "which requires a Linux host PC"
        ptree = choice.tree(dblu("p.all"), root="p")
        crawl(ptree)
        config.p = tree_prompt("Please select a platform", ptree)
        config._add_cfg("Platform", config.p.name)


    def _choose_ws():
        (ks,fws,pbs) = dblum(config.p.meta,["k","fw","pb"],config.p.meta,True)
        root = choice("<ws>", meta="ws")
        if (ks or fws) and HOST=="Linux":
            root.add(choice("Build from source", meta="bfs"))
        if pbs:
            root.add(choice("Use prebuilt configuration", meta="pbc"))
        config.ws = tree_prompt("Please specify whether you want to", root)
        config._add_cfg("Type", config.ws.name)


    def _choose_env():
        (ks, fws) = dblum(config.p.meta, ["k","fw"], config.p.meta, True)
        root = choice("<env>", meta="env")
        if ks:
            root.add(choice(dblu("k.name"), meta="k"))
        if fws:
            root.add(choice(dblu("fw.name"), meta="fw"))
        config.env = tree_prompt("Please select an environment", root)
        config._add_cfg("Environment", config.env.name)


    def _choose_fw():
        p = config.p.meta
        fwtree = choice.tree(dblu(p+".fw", p), "fw", p)
        config.fw = tree_prompt("Please select your firmware", fwtree)
        config._add_cfg("Firmware", config.fw.name)
        config._add_deps(config.fw.meta)
        config.manifest = dblu(config.fw.meta+".manifest", config.p.meta)


    def _choose_k():
        p = config.p.meta
        ktree = choice.tree(dblu(p+".k", p), "k", p)
        config.k = tree_prompt("Please select your kernel", ktree)
        config._add_cfg("Kernel",
            "{}{}".format(config.k.name,
                " ({})".format(config.k.descr) if config.k.descr else ""))
        config._add_deps(config.k.meta)
        config.manifest = dblu(config.k.meta+".manifest", config.p.meta)


    def _choose_fs():
        plat_fss = dblu(config.p.meta+".fs", config.p.meta)
        kernel_fss = dblu(config.k.meta+".fs", config.p.meta)
        fss = [fs for fs in kernel_fss if fs in plat_fss]
        fstree = choice.tree(fss, "fs", config.p.meta)
        config.fs = tree_prompt("Please select your filesystem", fstree)
        config._add_cfg("Filesystem", config.fs.name+" "+config.fs.descr)
        config._add_deps(config.fs.meta)


    def _choose_pb():
        choices = []
        for pb in dblu(config.p.meta+".pb", config.p.meta):
            choices.append(choice(dblu(pb+".name", config.p.meta), pb))
        if len(choices)==1:
            config.pb = choices[0]
            log.warn("only avail. option is "+config.pb.name)
        else:
            choices.append(choice("<< all >>", meta="all"))
            config.pb = prompt("Please select a configuration", choices)
        config._add_cfg("Configuration", config.pb.name)
        if config.pb.meta=="all":
            all_pb_deps = []
            for pb in dblu(config.p.meta+".pb", config.p.meta):
                pb_deps = dblu(pb+".deps", config.p.meta, noneAllowed=True)
                if pb_deps:
                    for d in pb_deps:
                        all_pb_deps.append(d)
            config.deps += list(set(all_pb_deps))
        else:
            config._add_deps(config.pb.meta)


def run():
    config.query()
    config.sync()
    return 0


def run_qa():
    check_empty_ws()
    script.start_qa()
    manifests = []
    for p in dblu("p.all"):
        log.info(">>> Running QA for platform "+p)
        (ks,fws,fss,pbs) = dblum(p,["k","fw","fs","pb"], p, noneAllowed=True)
        if ks:
            [manifests.append(dblu(k+".manifest", plat=p)) for k in ks]
        if fws:
            [manifests.append(dblu(fw+".manifest", plat=p)) for fw in fws]
        keys = (fss if fss else []) + (pbs if pbs else [])
        if keys:
            for k in keys:
                deps = dblu(k+".deps", plat=p, noneAllowed=True)
                if deps:
                    [sh.fetch(d, plat=p, force_fresh=True) for d in deps]
                    pass
    for m in list(set(manifests)):
        log.info(">>> Attempting to sync manifest "+m)
        sh.reposync(m, force_fresh=True)
    return script.end_qa()


if __name__ == "__main__":
    script.init()
    exit(run() if not script.qa_mode else run_qa())
